@@ -47,17 +47,23 @@ trap "rm -rf $LOCKFILE" EXIT
 PYENV=$BASE/pyenv
 
 
+# The file for 3.7b1 shipped with pyenv on Feb 6 2018
+# won't compile on Travis. So we use a forked version that
+# compiles openssl for us. We also beat them to the punch for 3.7b2, b3, b4
+# https://github.com/travis-ci/travis-ci/issues/9069
+
 if [ ! -d "$PYENV/.git" ]; then
   rm -rf $PYENV
-  git clone https://github.com/yyuu/pyenv.git $BASE/pyenv
+  git clone https://github.com/gevent/pyenv.git $BASE/pyenv
 else
   back=$PWD
   cd $PYENV
-  git fetch || echo "Update failed to complete. Ignoring"
+  git fetch || echo "Fetch failed to complete. Ignoring"
   git reset --hard origin/master
   cd $back
 fi
 
+
 SNAKEPIT=$BASE/snakepit
 
 install () {
@@ -66,16 +72,21 @@ install () {
   ALIAS="$2"
   mkdir -p $BASE/versions
   SOURCE=$BASE/versions/$ALIAS
+  OPENSSL_PATH=$SOURCE/openssl/lib
 
   if [ ! -e "$SOURCE" ]; then
     mkdir -p $SNAKEPIT
     mkdir -p $BASE/versions
-    $BASE/pyenv/plugins/python-build/bin/python-build $VERSION $SOURCE
+    LD_LIBRARY_PATH="$OPENSSL_PATH" $BASE/pyenv/plugins/python-build/bin/python-build $VERSION $SOURCE
   fi
  rm -f $SNAKEPIT/$ALIAS
  mkdir -p $SNAKEPIT
- $SOURCE/bin/python -m pip.__main__ install --upgrade pip wheel virtualenv
+ ls -l $SNAKEPIT
+ ls -l $BASE/versions
+ ls -l $SOURCE/
+ ls -l $SOURCE/bin
  ln -s $SOURCE/bin/python $SNAKEPIT/$ALIAS
+ LD_LIBRARY_PATH="$OPENSSL_PATH" $SOURCE/bin/python -m pip.__main__ install --upgrade pip wheel virtualenv
 }
 
 
@@ -85,28 +96,25 @@ for var in "$@"; do
       install 2.7.8 python2.7.8
       ;;
     2.7)
-      install 2.7.13 python2.7.13
-      ;;
-    3.2)
-      install 3.2.6 python3.2
-      ;;
-    3.3)
-      install 3.3.6 python3.3
+      install 2.7.14 python2.7.14
       ;;
     3.4)
-      install 3.4.5 python3.4.5
+      install 3.4.7 python3.4.7
       ;;
     3.5)
-      install 3.5.3 python3.5.3
+      install 3.5.4 python3.5.4
       ;;
     3.6)
-      install 3.6.0 python3.6.0
+      install 3.6.4 python3.6.4
+      ;;
+    3.7)
+      install 3.7.0b4 python3.7.0b4
       ;;
     pypy)
-      install pypy2-5.7.1 pypy571
+      install pypy2.7-5.10.0 pypy5100
       ;;
     pypy3)
-      install pypy3.5-5.7.1-beta pypy3.5_571
+      install pypy3.5-5.10.1 pypy3.5_5101
       ;;
   esac
 done
